module Tut09 where

{-T9.1-}
data NonEmptyList a = Single a | a `Cons` (NonEmptyList a)
  deriving (Show, Eq)

toList :: NonEmptyList a -> [a]
toList (Single x) = [x]
toList (x `Cons` xs) = x : toList xs

fromList :: [a] -> Maybe (NonEmptyList a)
fromList [] = Nothing
fromList (x : xs) = Just (go x xs)
  where go x [] = Single x
        go x (y : ys) = x `Cons` go y ys

nHead :: NonEmptyList a -> a
nHead (Single x) = x
nHead (x `Cons` _) = x

nTail :: NonEmptyList a -> Maybe (NonEmptyList a)
nTail (Single _) = Nothing
nTail (Cons _ t) = Just t

nAppend :: NonEmptyList a -> NonEmptyList a -> NonEmptyList a
(Single x) `nAppend` ys = x `Cons` ys
(x `Cons` xs) `nAppend` ys = x `Cons` xs `nAppend` ys

nTake :: Integer -> NonEmptyList a -> Maybe (NonEmptyList a)
nTake n _ | n <= 0 = Nothing
nTake n (Single x) 
  | n == 1 = Just $ Single x 
  | otherwise = Nothing
nTake n (x `Cons` xs)
  | n == 1 = Just $ Single x
  | otherwise = case nTake (n-1) xs of
      Nothing -> Nothing 
      Just res -> Just $ x `Cons` res

{-T9.3-}
type Name = String
data Polarity = Pos | Neg
  deriving Eq
data Literal = Literal Polarity Name
  deriving Eq
type Clause = [Literal]
type ConjForm = [Clause]

instance Show Polarity where
  show Pos = ""
  show Neg = "~"

instance Show Literal where
  show (Literal p a) = show p ++ a

lName :: Literal -> Name
lName (Literal _ n) = n

lPos :: Name -> Literal
lPos = Literal Pos

lNeg :: Name -> Literal
lNeg = Literal Neg

lIsPos :: Literal -> Bool
lIsPos (Literal p _) = p == Pos

lIsNeg :: Literal -> Bool
lIsNeg = not . lIsPos

lNegate :: Literal -> Literal
lNegate (Literal Pos n) = Literal Neg n
lNegate (Literal Neg n) = Literal Pos n

complements :: Literal -> Literal -> Bool
complements (Literal p n) (Literal p' n') = p /= p' && n == n'

type Valuation = [Name]

evalLiteral :: Valuation -> Literal -> Bool
evalLiteral val l@(Literal _ n)
  | lIsPos l  = n `elem` val
  | otherwise = n `notElem` val

evalClause :: Valuation -> Clause -> Bool
evalClause = any . evalLiteral

eval :: Valuation -> ConjForm -> Bool
eval = all . evalClause

clauseIsTauto :: Clause -> Bool
clauseIsTauto [] = False
clauseIsTauto (l:ls) = lNegate l `elem` ls || clauseIsTauto ls

